﻿using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

public class JointSimulator : MonoBehaviour
{
    #region constants
    private static readonly IReadOnlyList<Vector3> rotorPosition = new Vector3[4] {
        new Vector3( 0.6f, 0.25f,  0.6f),   // RF
        new Vector3( 0.6f, 0.25f, -0.6f),   // RB
        new Vector3(-0.6f, 0.25f, -0.6f),   // LB
        new Vector3(-0.6f, 0.25f,  0.6f)    // LF
    };

    public float DefaultThrottle = 24.5f;
    public Vector4 MaxGain = new Vector4(10f, 1f, 10f, 100f);
    public Vector4 KrP = new Vector4(  10f,   10f,   10f,   10f);
    public Vector4 KrI = new Vector4( 0.2f,  0.2f,  0.2f,    0f);
    public Vector4 KrD = new Vector4(1000f, 1000f, 1000f, 1000f);
    #endregion

    #region public accessor
    public bool Flight { get => onflight; set => onflight = value; }    // takeoff T, land F

    public float Pitch    { get => axis.x; set => axis.x = value; }     // front +, back  - (theta)
    public float Yaw      { get => axis.y; set => axis.y = value; }     // cw    +, ccw   - (psi)
    public float Roll     { get => axis.z; set => axis.z = value; }     // left  +, right - (phi)
    public float Throttle { get => axis.w; set => axis.w = value; }     // up velocity

    public Vector3 Accelero => accelero;    // accelerometer navdata
    public Vector3 Rotation => rotation;    // gyroscope + magnetometer navdata
    public float   Altitude => altitude;    // ultrasound + pressure navdata
    #endregion

    #region private attributes
    private new Rigidbody rigidbody;
    private new Transform transform;

    [SerializeField] private bool onflight;
    [SerializeField] private Vector4 axis = new Vector4(0, 0, 0, 98f);

    [SerializeField] private Vector3 accelero;
    [SerializeField] private Vector3 rotation;
    [SerializeField] private float   altitude;
    #endregion

    #region inspector view variable
    private Vector3 prevVelocity, velocity;

    [SerializeField] private float[] rotorThrust = new float[4]; // RF, RB, LB, LF
    [SerializeField] private float   rotorTorque = 0f;

    [SerializeField] private Vector4 pAxis, iAxis, dAxis, prevPAxis;
    [SerializeField] private Vector4 accAxis, prevAccAxis;
    #endregion

    void Start() {
        rigidbody = GetComponent<Rigidbody>();
        transform = GetComponent<Transform>();
    }

    void FixedUpdate() {
        prevVelocity = velocity;
        velocity = transform.InverseTransformDirection(rigidbody.velocity);
        accelero = (velocity - prevVelocity) / Time.fixedDeltaTime;
        rotation = transform.eulerAngles;
        altitude = transform.position.y;

        var targetRotation = Quaternion.Euler(axis);
        var currentRotation = transform.rotation;
        var deltaRotation = Quaternion.Inverse(currentRotation) * targetRotation;

        prevPAxis = pAxis;
        pAxis = deltaRotation.eulerAngles;
        pAxis.w = axis.w - altitude;
        if (pAxis.x >= 180f) pAxis.x -= 360f;
        if (pAxis.y >= 180f) pAxis.y -= 360f;
        if (pAxis.z >= 180f) pAxis.z -= 360f;
        iAxis += Vector4.Scale(KrI, pAxis);
        dAxis = pAxis - prevPAxis;

        prevAccAxis = accAxis;
        accAxis = Vector4.Scale(KrP, pAxis) + Vector4.Scale(KrI, iAxis) + Vector4.Scale(KrD, dAxis);
        accAxis.w += DefaultThrottle;

        if (Mathf.Abs(accAxis.x - prevAccAxis.x) >= MaxGain.x)
            accAxis.x = prevAccAxis.x + Mathf.Clamp(accAxis.x - prevAccAxis.x, -MaxGain.x, MaxGain.x);
        if (Mathf.Abs(accAxis.y - prevAccAxis.y) >= MaxGain.y)
            accAxis.y = prevAccAxis.y + Mathf.Clamp(accAxis.y - prevAccAxis.y, -MaxGain.y, MaxGain.y);
        if (Mathf.Abs(accAxis.z - prevAccAxis.z) >= MaxGain.z)
            accAxis.z = prevAccAxis.z + Mathf.Clamp(accAxis.z - prevAccAxis.z, -MaxGain.z, MaxGain.z);
        if (Mathf.Abs(accAxis.w - prevAccAxis.w) >= MaxGain.w)
            accAxis.w = prevAccAxis.w + Mathf.Clamp(accAxis.w - prevAccAxis.w, -MaxGain.w, MaxGain.w);

        rotorThrust[0] = accAxis.w - accAxis.x + accAxis.y + accAxis.z;
        rotorThrust[1] = accAxis.w + accAxis.x - accAxis.y + accAxis.z;
        rotorThrust[2] = accAxis.w + accAxis.x + accAxis.y - accAxis.z;
        rotorThrust[3] = accAxis.w - accAxis.x - accAxis.y - accAxis.z;
        rotorTorque = rotorThrust[0] - rotorThrust[1] + rotorThrust[2] - rotorThrust[3];

        // apply thrust
        rotorPosition
            .Select(p => transform.TransformPoint(p))   // local to world position
            .Zip(rotorThrust, (p, f) => new { p, f })
            .ToList()
            .ForEach(e => rigidbody.AddForceAtPosition(transform.up * e.f, e.p));   // rotor force

        rigidbody.AddRelativeTorque(new Vector3(0, rotorTorque, 0));
    }
}
