﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public class MoveController : MonoBehaviour
{
    public enum DroneState { Inactive, TakeOff, OnFlight, Landing };

    public bool Emergency {     // emergency flag
        get => emergency;
        set {
            if (emergency != value) {
                emergency = value;
                thrust.Emergency = emergency;
                if (!emergency) {
                    attitude.reset();
                }
            }
        }
    }

    public DroneState Status => droneState;
    public bool Flight {
        set {
            if (value) {
                if (droneState == DroneState.Inactive)
                    droneState = DroneState.TakeOff;
            }
            else {
                if (droneState == DroneState.OnFlight || droneState == DroneState.TakeOff)
                    droneState = DroneState.Landing;
            }
        }
    }

    public float Front { get => movement.x / 5f; set { movement.x = value * 5f; reset(); } }  // front-back movement
    public float Up    { get => movement.y / 5f; set { movement.y = value * 5f; reset(); } }  //    up-down movement
    public float Left  { get => movement.z / 5f; set { movement.z = value * 5f; reset(); } }  // left-right movement
    public float Turn  { get => movement.w     ; set { movement.w = value     ; reset(); } }  // turn(+ccw) movement

    [SerializeField] [Readonly] private bool emergency = false;
    [SerializeField] private DroneState droneState = DroneState.Inactive;
    [SerializeField] private Vector4 movement;

    [Header("PID Control values")]
    [SerializeField] private Vector4 KrP = new Vector4(5f, 8f, 5f, 200f);
    [SerializeField] private Vector4 KrI = new Vector4(0.1f, 0.4f, 0.1f, 0.05f);
    [SerializeField] private Vector4 KrD = new Vector4(200f, 100f, 200f, 2000f);
    [SerializeField] [Readonly] private Vector4 pMove, iMove, dMove;
    private Vector4 prevPMove;
    [SerializeField] [Readonly] private Vector4 force;

    [SerializeField] [Readonly] private float antigravity;
    [SerializeField] [Readonly] private Vector4 targetVelocity, currentVelocity;
    private float prevYaw, yaw;

    private NavdataHandler     navdata;
    private AttitudeController attitude;
    private ThrustController   thrust;

    void Start() {
        navdata  = GetComponent<NavdataHandler>();
        attitude = GetComponent<AttitudeController>();
        thrust   = GetComponent<ThrustController>();
    }

    void FixedUpdate() {
        prevYaw = yaw;
        yaw = navdata.Rotation.y;

        currentVelocity = navdata.Velocity;
        currentVelocity.w = (yaw - prevYaw).RelativeAngle();

        switch (droneState) {
            case DroneState.TakeOff:
                if (currentVelocity.y > 0.5f) {
                    droneState = DroneState.OnFlight;
                    goto case DroneState.OnFlight;
                }
                antigravity += 0.05f;

                force = Vector4.zero;
                force.y += antigravity;
                break;

            case DroneState.OnFlight:
                targetVelocity = movement;

                prevPMove = pMove;
                pMove = targetVelocity - currentVelocity;
                pMove.w = pMove.w.RelativeAngle();
                iMove += Vector4.Scale(KrI, pMove);
                dMove = pMove - prevPMove;

                force = Vector4.Scale(KrP, pMove) + Vector4.Scale(KrI, iMove) + Vector4.Scale(KrD, dMove);
                force.y += antigravity;
                break;

            case DroneState.Landing:
                targetVelocity = Vector4.zero;
                if (navdata.Altitude > 0.2f) {
                    targetVelocity.y = -4f;

                    prevPMove = pMove;
                    pMove = targetVelocity - currentVelocity;
                    pMove.w = pMove.w.RelativeAngle();
                    iMove += pMove * Time.fixedDeltaTime;
                    dMove = (pMove - prevPMove) / Time.fixedDeltaTime;

                    force = Vector4.Scale(KrP, pMove) + Vector4.Scale(KrI, iMove) + Vector4.Scale(KrD, dMove);
                    force.y += antigravity;
                }
                else {
                    reset();
                    antigravity = 0f;
                    force = Vector4.zero;
                    droneState = DroneState.Inactive;
                    goto case DroneState.Inactive;
                }
                break;

            case DroneState.Inactive:
                break;
        }

        attitude.Roll  = Mathf.Clamp(-force.x, -60f, 60f);
        thrust.Thrust  = force.y;
        attitude.Pitch = Mathf.Clamp( force.z, -60f, 60f);
        attitude.Yaw  += Mathf.Clamp(currentVelocity.w + force.w * Time.fixedDeltaTime, -10f, 10f);
        Debug.Log(currentVelocity.w + force.w * Time.deltaTime);
    }

    public void reset() => pMove = iMove = dMove = prevPMove = Vector3.zero;
}
